/*
 * Copyright (C) 2003 BLADOX, s.r.o.  All rights reserved.
 * 
 * This file is part of an example program for Turbo. This example
 * program may be used, distributed and modified without limitation.
 *
 */


/* Implements ECC over Z/pZ for curve y^2 = x^3 - 3x + b
 *
 * All curves taken from NIST recommendation paper of July 1999
 * Available at http://csrc.nist.gov/cryptval/dss.htm
 */

#include <config.h>
#include <turbo/turbo.h>

#include <string.h>

#include "certs.h"
#include "mycrypt.h"

u8   PROGMEM ecc192_prime[]="6277101735386680763835789423207666416083908700390324961279";
u8   PROGMEM ecc192_B[]="64210519e59c80e70fa7e9ab72243049feb8deecc146b9b1";
u8   PROGMEM ecc192_order[]="6277101735386680763835789423176059013767194773182842284081";
u8   PROGMEM ecc192_Gx[]="188da80eb03090f67cbf20eb43a18800f4ff0afd82ff1012";
u8   PROGMEM ecc192_Gy[]="07192b95ffc8da78631011ed6b24cdd573f977a11e794811";

/*
char PROGMEM ecc256_prime[]="115792089210356248762697446949407573530086143415290314195533631308867097853951";
char PROGMEM ecc256_B[]="5ac635d8aa3a93e7b3ebbd55769886bc651d06b0cc53b0f63bce3c3e27d2604b";
char PROGMEM ecc256_order[]="115792089210356248762697446949407573529996955224135760342422259061068512044369";
char PROGMEM ecc256_Gx[]="6b17d1f2e12c4247f8bce6e563a440f277037d812deb33a0f4a13945d898c296";
char PROGMEM ecc256_Gy[]="4fe342e2fe1a7f9b8ee7eb4a7c0f9e162bce33576b315ececbb6406837bf51f5";
*/

void out_mp(mp_int *a)
{
	unsigned i;
	dbsp("MP_SIGN ");dbch(SIGN(a));dbc('\n');
	dbsp("MP_USED ");dbch(USED(a));dbc('\n');
	dbsp("MP_ALLOC ");dbch(ALLOC(a));dbc('\n');
	for(i=0;i<USED(a);i++){dbih(DIGIT(a,i));dbc(' ');}
	dbc('\n');
	db_imalloc();
}

void out_mp_8(mp_int *a)
{
	unsigned i;
	unsigned char *c=DIGITS(a);
	dbsp("MP_SIGN ");dbch(SIGN(a));dbc('\n');
	dbsp("MP_USED ");dbch(USED(a));dbc('\n');
	dbsp("MP_ALLOC ");dbch(ALLOC(a));dbc('\n');
	for(i=0;i<USED(a)*2;i++){dbch(c[i]);dbc(' ');}
	dbc('\n');
	db_imalloc();
}

void out_mp_8e(mp_int *a)
{
	unsigned i;
	unsigned char *c=rw(&a->dp);
	unsigned char u=rb(&a->used);
	dbsp("MP_SIGN ");dbch(rb(&a->sign));dbc('\n');
	dbsp("MP_USED ");dbch(rb(&a->used));dbc('\n');
	dbsp("MP_ALLOC ");dbch(rb(&a->alloc));dbc('\n');
	for(i=0;i<u*2;i++){dbch(rb(c+i));dbc(' ');}
	dbc('\n');
	db_imalloc();
}



/*
static ecc_point *new_point(void)
{
   ecc_point *p;
   p = imalloc(sizeof(ecc_point));
   db_imalloc();
   if (p == NULL) {
      return NULL;
   }
   if (mp_init_multi(&p->x, &p->y, NULL) != MP_OKAY) {
      ifree(p);
      return NULL;
   }
   return p;
}

static void del_point(ecc_point *p)
{
   mp_clear_multi(&p->x, &p->y, NULL);
   ifree(p);
   db_imalloc();
}
*/

/* double a point R = 2P, R can be P*/
//static int dbl_point(ecc_point *P, ecc_point *R, mp_int *modulus)
static int dbl_point(mp_int *Px,mp_int *Py, mp_int *Rx, mp_int *Ry, mp_int *modulus)
{
   mp_int s, tmp, tmpx;
   int res;

//   dbsp("DBP 1\n");
   if (mp_init_multi(&s, &tmp, &tmpx, NULL) != MP_OKAY) { 
    //  crypt_error = "Out of memory in ecc.c::dbl_point().";
      return CRYPT_ERROR;
   }
//      dbsp("DBL_POINT_1\n");out_mp(&P->x);
//      dbsp("DBL_POINT_1\n");out_mp(&P->y);


   /* s = (3Xp^2 + a) / (2Yp) */
   if (mp_mul_2(Py, &tmp) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_1_1\n");out_mp(&tmp);
   if (mp_invmod(&tmp, modulus, &tmp) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_1_2\n");out_mp(&tmp);
   if (mp_sqrmod(Px, modulus, &s) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_1_3\n");out_mp(&s);
   if (mp_mul_d(&s, 3, &s) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_1_4\n");out_mp(&s);
   if (mp_sub_d(&s, 3, &s) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_1_5\n");out_mp(&s);
   if (mp_mulmod(&s, &tmp, modulus, &s) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_2\n");out_mp(&s);

   /* Xr = s^2 - 2Xp */
   if (mp_sqrmod(&s, modulus, &tmpx) != MP_OKAY) { goto error; }
   if (mp_sub(&tmpx, Px, &tmpx) != MP_OKAY) { goto error; }
   if (mp_submod(&tmpx, Px, modulus, &tmpx) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_3\n");out_mp(&tmpx);

   /* Yr = -Yp + s(Xp - Xr)  */
   if (mp_sub(Px, &tmpx, &tmp) != MP_OKAY) { goto error; }
   if (mp_mul(&tmp, &s, &tmp) != MP_OKAY) { goto error; }
   if (mp_submod(&tmp, Py, modulus, Ry) != MP_OKAY) { goto error; }
   if (mp_copy(&tmpx, Rx) != MP_OKAY) { goto error; }
//   dbsp("DBL_POINT_4\n");out_mp(&R->x);
//   dbsp("DBL_POINT_4\n");out_mp(&R->y);

   res = CRYPT_OK;
   goto done;
error:
   res = CRYPT_ERROR;
done:
   mp_clear_multi(&tmpx, &tmp, &s, NULL);
   return res;
}

/* add two different points over Z/pZ, R = P + Q, note R can equal either P or Q */
//static int add_point(ecc_point *P, ecc_point *Q, ecc_point *R, mp_int *modulus)
static int add_point(mp_int *Px, mp_int *Py, mp_int *Qx, mp_int *Qy, mp_int *Rx, mp_int *Ry, mp_int *modulus)
{
   mp_int s, tmp, tmpx;
   int res;

//   dbsp("ADD_POINT_1\n");out_mp(&P->x);
//   dbsp("ADD_POINT_1\n");out_mp(&P->y);
//   dbsp("ADD_POINT_2\n");out_mp(&Q->x);
//   dbsp("ADD_POINT_2\n");out_mp(&Q->y);

   if (mp_init(&tmp) != MP_OKAY) { 
      return CRYPT_ERROR;
   }

   /* is P==Q or P==-Q? */
   mp_neg(Qy, &tmp);
   mp_mod(&tmp, modulus, &tmp); //!!!optimize
   if (!mp_cmp(Px, Qx))
      if (!mp_cmp(Py, Qy) || !mp_cmp(Py, &tmp)) {
         mp_clear(&tmp);
         return dbl_point(Px, Py, Rx, Ry, modulus);
      }

   if (mp_init_multi(&tmpx, &s, NULL) != MP_OKAY) { 
      mp_clear(&tmp);
      return CRYPT_ERROR;
   }

   /* get s = (Yp - Yq)/(Xp-Xq) mod p */
   if (mp_submod(Px, Qx, modulus, &tmp) != MP_OKAY) { goto error; }
   if (mp_invmod(&tmp, modulus, &tmp) != MP_OKAY) { goto error; }
   if (mp_submod(Py, Qy, modulus, &s) != MP_OKAY) { goto error; }
   if (mp_mulmod(&s, &tmp, modulus, &s) != MP_OKAY) { goto error; }
//   dbsp("ADD_POINT_3\n");out_mp(&s);

   /* Xr = s^2 - Xp - Xq */
   if (mp_sqrmod(&s, modulus, &tmp) != MP_OKAY) { goto error; }
   if (mp_sub(&tmp, Px, &tmp) != MP_OKAY) { goto error; }
   if (mp_sub(&tmp, Qx, &tmpx) != MP_OKAY) { goto error; }
//   dbsp("ADD_POINT_4\n");out_mp(&tmpx);

   /* Yr = -Yp + s(Xp - Xr) */
   if (mp_sub(Px, &tmpx, &tmp) != MP_OKAY) { goto error; }
   if (mp_mul(&tmp, &s, &tmp) != MP_OKAY) { goto error; }
   if (mp_submod(&tmp, Py, modulus, Ry) != MP_OKAY) { goto error; }
   if (mp_mod(&tmpx, modulus, Rx) != MP_OKAY) { goto error; }
//   dbsp("ADD_POINT_5\n");out_mp(&R->x);

   res = CRYPT_OK;
   goto done;
error:
   res = CRYPT_ERROR;
done:
   mp_clear_multi(&s, &tmpx, &tmp, NULL);
   return res;
}

#if 0
/* perform R = kG where k == integer and G == ecc_point */
static int my_ecc_mulmod(mp_int *k, ecc_point *R, mp_int *modulus)
{
   ecc_point *tG;
   int i, j, z, first, res;
   mp_digit d;

   i=USED(k)*16;
   
   /* make a copy of G incase R==G */
   tG = new_point();
   if (tG == NULL) { 
      return CRYPT_ERROR;
   }

   /* tG = G */
   if (mp_read_radix(&tG->x, ecc192_Gx, 16) != MP_OKAY) { goto error; }
   if (mp_read_radix(&tG->y, ecc192_Gy, 16) != MP_OKAY) { goto error; }

   /* set result to G, R = G */
   if (mp_read_radix(&R->x, ecc192_Gx, 16) != MP_OKAY) { goto error; }
   if (mp_read_radix(&R->y, ecc192_Gy, 16) != MP_OKAY) { goto error; }

   first = 0;

   dbsp("EMUL_0\n");
   out_mp_8(&R->x);
   out_mp_8(&R->y);
   out_mp_8(k);
   /*
   dbsp("EMUL_0_0\n");
   out_mp(&R->x);
   out_mp(&R->y);
   out_mp(k);
   */

   /* now do dbl+add through all the bits */
   for (j = i-1; j >= 0; j--) {
	   dbsp("EMUL: ");dbch(j);dbc('\n');
	   db_imalloc();
       if (first)
	    {
           if (dbl_point(R, R, modulus) == CRYPT_ERROR) {
		   goto error; }
//	   dbsp("EMUL_1\n");out_mp(&R->x);
//	   dbsp("EMUL_1\n");out_mp(&R->y);
	    }
//       if (bits[j] == 1) {
	if (DIGIT(k,j/16)&(1 << j%16)) {
          if (first)
		{
             if (add_point(R, tG, R, modulus) == CRYPT_ERROR) { 
		     goto error; }
//	   dbsp("EMUL_2\n");out_mp(&R->x);
//	   dbsp("EMUL_2\n");out_mp(&R->y);
		}
          first = 1;
       }
   }
   
   dbsp("EMUL_3\n");
   out_mp_8(&R->x);
   out_mp_8(&R->y);
   out_mp_8(k);
   /*
   dbsp("EMUL_4\n");
   out_mp(&R->x);
   out_mp(&R->y);
   out_mp(k);
   */
   res = CRYPT_OK;
   goto done;
error:
   res = CRYPT_ERROR;
done:
   del_point(tG);
   return res;
}
#endif

//#ifdef TEST_VECT
static u8 PROGMEM buf_r[]=
{ 
	0x00, 0xCE, 0xBC, 0x32, 0x37, 0x96, 0xA8, 0x6B, 0x27, 0x00, 
	0xFC, 0x7E, 0x36, 0xD6, 0x01, 0x97, 0xF7, 0x29, 0x92, 0x8E, 
	0x60, 0x3D, 0xED, 0x25, 0xAF, 0x83, 0x5B, 0x54, 0x8D, 0x61, 
	0x64, 0xF5, 0xEF, 0x50, 0x56, 0x43, 0x40, 0x9A, 0xFF, 0x37, 
	0x82, 0x83, 0xB1, 0x57, 0x16, 0xBD, 0x94, 0x85, 0x8B, 0x00, 
	0x93, 0x8D, 0x27, 0x75, 0x7B, 0xDD, 0xC0, 0x64, 0xF2, 0xBF, 
	0xE4, 0xCE, 0xD2, 0xF4, 0x18, 0xD0
};
//#endif

extern unsigned char keys_clear;

void ecc_clear_keys()
{
unsigned char l;
ecc_key *ekey;
for(l=0;l<nr_ecc_keys;l++)
{
dbsp("KEY: ");dbch(l);dbc('\n');
ekey=ecc_keys+l*sizeof(ecc_key);
wb(&ekey->ekey_status,KEY_START);
wb(&ekey->ekey_ack,KEY_START);
}
keys_clear=0;
reg_action(ACTION_IDLE_TASK);
}

void certs_idle_task_2()
{
ecc_key *ekey;
mp_int prime;
mp_int tGx, tGy;
mp_int Rx, Ry;
mp_int k;
unsigned char l;
int i,j;
unsigned char first;
unsigned char ekey_status;
unsigned char ekey_ack;
unsigned char ekey_first;
unsigned char ekey_j;
unsigned char ekey_first_ack;
unsigned char ekey_j_ack;
//unsigned char compress_y;

#define used_k	(0x0C*16)

if(keys_clear==1)
	ecc_clear_keys();

//while(1)
//{
dbsp("IDLE_TASK_1\n");
init_imalloc();

if (mp_init(&Rx) != MP_OKAY)  goto error;
if (mp_init(&Ry) != MP_OKAY)  goto error;
if (mp_init(&k) != MP_OKAY)  goto error;
if (mp_init(&tGx) != MP_OKAY)  goto error;
if (mp_init(&tGy) != MP_OKAY)  goto error;
if (mp_init(&prime) != MP_OKAY) goto error;

for(l=0;l<nr_ecc_keys;l++)
{
dbsp("KEY: ");dbch(l);dbc('\n');
ekey=ecc_keys+l*sizeof(ecc_key);

ekey_status=rb(&ekey->ekey_status);
ekey_ack=rb(&ekey->ekey_ack);
ekey_first=rb(&ekey->ekey_first);
ekey_j=rb(&ekey->ekey_j);
ekey_first_ack=rb(&ekey->ekey_first_ack);
ekey_j_ack=rb(&ekey->ekey_j_ack);

dbsp("IDLE_TASK_2\n");
if(!(ekey_status==KEY_DONE && ekey_status == ekey_ack))
{
dbsp("IDLE_TASK_3\n");
if(!(ekey_status==KEY_COMPRESS && ekey_status ==ekey_ack))
{
if (mp_read_radix(&prime, ecc192_prime, 10) != MP_OKAY) goto error;
if (mp_read_radix(&tGx, ecc192_Gx, 16) != MP_OKAY)  goto error; 
if (mp_read_radix(&tGy, ecc192_Gy, 16) != MP_OKAY)  goto error; 

if(ekey_status==KEY_START || ekey_status !=ekey_ack)
{
dbsp("KEY_INIT\n");

dbsp("PX\n");	out_mp_8e(&ekey->x); 
dbsp("PY\n");	out_mp_8e(&ekey->y); 
dbsp("KEY_INIT_1\n");

// make rnd
// to eeprom
//#ifdef TEST_VECT
for(i=0;i<KEYSIZE+1;i++)wb(&ekey->k[i],rb(buf_r+i)); 
//#else
//#endif

wb(&ekey->ekey_status,KEY_INITED);
wb(&ekey->ekey_first,0);
wb(&ekey->ekey_j,used_k); 
wb(&ekey->ekey_first_ack,0);
wb(&ekey->ekey_j_ack,used_k);
wb(&ekey->ekey_ack,KEY_INITED);

ekey_status=KEY_INITED;
ekey_ack=KEY_INITED;
ekey_first=0;
ekey_first_ack=0;
ekey_j=used_k;
ekey_j_ack=used_k;
}

{
//#ifdef TEST_VECT
u8 * bf=imalloc(sizeof(buf_r));
memcpy(bf,buf_r, sizeof(buf_r));
//#endif
if (mp_read_raw(&k, bf, KEYSIZE+1) != MP_OKAY) goto error;
ifree(bf);
}

dbsp("KEY_INIT_2\n");
if(ekey_j!=ekey_j_ack || ekey_first!=ekey_first_ack)
{
dbsp("KEY_REINIT\n");
ekey_first=0;
ekey_j=used_k;
}

i=ekey_j-1;
first=ekey_first;

dbsp("KEY_INIT_3\n");
dbsp("I: ");dbch(i);dbc('\n');

if(i != used_k-1)
{
dbsp("KEY_INIT_4\n");
e_mp_copy(&ekey->x, &Rx); 
e_mp_copy(&ekey->y, &Ry);
}
else
{
mp_copy(&tGx,&Rx);
mp_copy(&tGy,&Ry);
}

for (j = i; j >= 0; j--) {
	dbsp("EMUL: ");dbch(j);dbc('\n');
//	db_imalloc();
	if (first) 
		if (dbl_point(&Rx, &Ry, &Rx, &Ry, &prime) == CRYPT_ERROR)  goto error; 
	
		if (DIGIT(&k,j/16)&(1 << j%16)) { 
			if (first)
				if (add_point(&Rx, &Ry, &tGx, &tGy, &Rx, &Ry, &prime) == CRYPT_ERROR) goto error; 
			first = 1;
		}

// FULL SPEED	10
// NOKIA	5
#define KEY_STORE_PERIOD	5
	if(j%KEY_STORE_PERIOD==0)
	{
		dbsp("KEY_STORED\n");
		wb(&ekey->ekey_j,j);
		wb(&ekey->ekey_first,first);
		mp_copy_e(&Rx, &ekey->x);
		mp_copy_e(&Ry, &ekey->y);
		wb(&ekey->ekey_first_ack,first);
		wb(&ekey->ekey_j_ack,j);
		dbsp("KEY_STORED_6\n");
	}
}
	wb(&ekey->ekey_j,j);
	wb(&ekey->ekey_first,first);

	mp_copy_e(&Rx, &ekey->x);
	mp_copy_e(&Ry, &ekey->y);
	
	wb(&ekey->ekey_first_ack,first);
	wb(&ekey->ekey_j_ack,j);

	wb(&ekey->ekey_status,KEY_COMPRESS);
	wb(&ekey->ekey_ack,KEY_COMPRESS);
	dbsp("KEY_COMPRESS\n");

	dbsp("PX\n");	out_mp_8e(&ekey->x); 
	dbsp("PY\n");	out_mp_8e(&ekey->y); 
	}
else
{
dbsp("KEY_INIT_COMPRESS\n");
e_mp_copy(&ekey->x, &Rx); 
e_mp_copy(&ekey->y, &Ry);
}

	dbsp("KEY_COMPRESS_1\n");

   /* get x^3 - 3x + b */
   if (mp_read_radix(&k, ecc192_B, 16) != MP_OKAY) { goto error; }
//   dbsp("COMP_2\n");out_mp_8(&k);
   if (mp_expt_d(&Rx, 3, &tGx) != MP_OKAY) { goto error; }
//   dbsp("COMP_3\n");out_mp_8(&tGx);
   if (mp_mul_d(&Rx, 3, &tGy) != MP_OKAY) { goto error; }
//   dbsp("COMP_4\n");out_mp_8(&tGy);
   if (mp_sub(&tGx, &tGy, &tGx) != MP_OKAY) { goto error; }
//   dbsp("COMP_5\n");out_mp_8(&tGx);
   if (mp_add(&tGx, &k, &tGx) != MP_OKAY) { goto error; }
//   dbsp("COMP_6\n");out_mp_8(&tGx);
   if (mp_read_radix(&k, ecc192_prime, 10) != MP_OKAY) { goto error; }
//   dbsp("COMP_7\n");out_mp_8(&k);
   if (mp_mod(&tGx, &k, &tGx) != MP_OKAY) { goto error; }
//   dbsp("COMP_8\n");out_mp_8(&tGx);

   /* now find square root */
   if (mp_add_d(&k, 1, &tGy) != MP_OKAY) { goto error; }
//   dbsp("COMP_9\n");out_mp_8(&tGy);
   if (mp_div_2(&tGy, &tGy) != MP_OKAY) { goto error; }
//   dbsp("COMP_10\n");out_mp_8(&tGy);
   if (mp_div_2(&tGy, &tGy) != MP_OKAY) { goto error; }              /* tGy = (p+1)/4 */
//   dbsp("COMP_11\n");out_mp_8(&tGy);
   if (mp_exptmod(&tGx, &tGy, &k, &tGx) != MP_OKAY) { goto error; }   /* tGx  = (x^3 - 3x + b)^((p+1)/4) mod p */
//   dbsp("COMP_12\n");out_mp_8(&tGx);

   /* if tmp equals the y point give a 0, otherwise 1 */
if (mp_cmp(&tGx, &Ry) == 0)
{
	wb(&ekey->compress_y,0);
	dbsp("COMPRESS_0\n");
}
else
{
	wb(&ekey->compress_y,1);
	dbsp("COMPRESS_1\n");
}

 	wb(&ekey->ekey_status,KEY_DONE);
	wb(&ekey->ekey_ack,KEY_DONE);

	dbsp("KEY_DONE\n");
}
}

	unreg_action(ACTION_IDLE_TASK);
	goto sleep;
error:
	dbsp("ERROR\n");
	unreg_action(ACTION_IDLE_TASK);
sleep:
	return;
}

#if 0
int ecc_make_key(int keysize, ecc_key *key)
{
   unsigned char buf[33];
   int x, res,i ;
   mp_int prime;

   keysize = 24;
   x= 0;

   /* make up random string */
   buf[0] = 0;
#if 0
   if (prng_descriptor[wprng].read(buf+1, keysize, prng) != (unsigned long)keysize) {
      //crypt_error = "Error reading PRNG in ecc_make_key().";
      return CRYPT_ERROR;
   }
#endif
   for(i=0;i<keysize+1;i++){buf[i]=buf_r[i];}
//   dbsp("X1\n");
   
   /* setup the key variables */
   
   if (mp_init(&key->pubkey.x) != MP_OKAY)  return CRYPT_ERROR;
   if (mp_init(&key->pubkey.y) != MP_OKAY)  return CRYPT_ERROR;
   if (mp_init(&key->k) != MP_OKAY)  return CRYPT_ERROR;
   if (mp_init(&prime) != MP_OKAY)  return CRYPT_ERROR;

//   dbsp("KEY\n"); out_mp_8(&key->k);
   /*
   if (mp_init_multi(&key->pubkey.x, &key->pubkey.y, &key->k, &prime, NULL) != MP_OKAY) { 
      //crypt_error = "Out of memory in ecc_make_key()."; 
      return CRYPT_ERROR;
   }
   */
//   dbsp("X2\n");
   /* read in the specs for this key */
   if (mp_read_radix(&prime, ecc192_prime, 10) != MP_OKAY) { goto error; }
//   out_mp_8(&prime);
//   dbsp("X3\n");
//   dbsp("KS "); dbih(keysize);dbc('\n');
//   out_mp_8(&key->k);
   if (mp_read_raw(&key->k, buf, keysize+1) != MP_OKAY) { goto error; }
//   dbsp("X5\n");

   /* make the public key */
   if (my_ecc_mulmod(&key->k, &key->pubkey, &prime) == CRYPT_ERROR) { goto error; }
//   key->type = PK_PRIVATE;
   /*
   dbsp("KEY\n");
   out_mp(&key->pubkey.x);
   out_mp(&key->pubkey.y);
   */
   
   /* free up ram */
   res = CRYPT_OK;
   goto done;
error:
//   dbsp("X6\n");
   res = CRYPT_ERROR;
done:
//   dbsp("X7\n");
//   db_malloc();
   mp_clear(&prime);
   return res;
}

void ecc_free(ecc_key *key)
{
   mp_clear_multi(&key->pubkey.x, &key->pubkey.y, &key->k, NULL);
}

int compress_y_point(ecc_point *pt, int *result)
{
   mp_int tmp, tmp2, p;
   int res;

//   dbsp("COMP_1\n");
//   db_malloc();
   if (mp_init_multi(&tmp, &tmp2, &p, NULL) != MP_OKAY) {
      //crypt_error = "Out of memory in compress_y_point().";
      return CRYPT_ERROR;
   }

   /* get x^3 - 3x + b */
   if (mp_read_radix(&p, ecc192_B, 16) != MP_OKAY) { goto error; }
//   dbsp("COMP_2\n");out_mp(&p);
   if (mp_expt_d(&pt->x, 3, &tmp) != MP_OKAY) { goto error; }
//   dbsp("COMP_3\n");out_mp(&tmp);
   if (mp_mul_d(&pt->x, 3, &tmp2) != MP_OKAY) { goto error; }
//   dbsp("COMP_4\n");out_mp(&tmp2);
   if (mp_sub(&tmp, &tmp2, &tmp) != MP_OKAY) { goto error; }
//   dbsp("COMP_5\n");out_mp(&tmp);
   if (mp_add(&tmp, &p, &tmp) != MP_OKAY) { goto error; }
//   dbsp("COMP_6\n");out_mp(&tmp);
   if (mp_read_radix(&p, ecc192_prime, 10) != MP_OKAY) { goto error; }
//   dbsp("COMP_7\n");out_mp(&p);
   if (mp_mod(&tmp, &p, &tmp) != MP_OKAY) { goto error; }
//   dbsp("COMP_8\n");out_mp(&tmp);

   /* now find square root */
   if (mp_add_d(&p, 1, &tmp2) != MP_OKAY) { goto error; }
//   dbsp("COMP_9\n");out_mp(&tmp2);
   if (mp_div_2(&tmp2, &tmp2) != MP_OKAY) { goto error; }
//   dbsp("COMP_10\n");out_mp(&tmp2);
   if (mp_div_2(&tmp2, &tmp2) != MP_OKAY) { goto error; }              /* tmp2 = (p+1)/4 */
//   dbsp("COMP_11\n");out_mp(&tmp2);
   if (mp_exptmod(&tmp, &tmp2, &p, &tmp) != MP_OKAY) { goto error; }   /* tmp  = (x^3 - 3x + b)^((p+1)/4) mod p */
//   dbsp("COMP_12\n");out_mp(&tmp);

   /* if tmp equals the y point give a 0, otherwise 1 */
   if (mp_cmp(&tmp, &pt->y) == 0)
      *result = 0;
   else
      *result = 1;
   
   res = CRYPT_OK;
   goto done;
error:
   res = CRYPT_ERROR;
   //crypt_error = "Out of memory in compress_y_point().";
done:
   mp_clear_multi(&p, &tmp, &tmp2, NULL);
   return res;
}
#endif

#if 0
int expand_y_point(ecc_point *pt, int idx, int result)
{
   mp_int tmp, tmp2, p;
   int res;
 
   if (mp_init_multi(&tmp, &tmp2, &p, NULL) != MP_OKAY) {
      //crypt_error = "Out of memory in compress_y_point().";
      return CRYPT_ERROR;
   }

   /* get x^3 - 3x + b */
   if (mp_read_radix(&p, sets[idx].B, 16) != MP_OKAY) { goto error; }
   if (mp_expt_d(&pt->x, 3, &tmp) != MP_OKAY) { goto error; }
   if (mp_mul_d(&pt->x, 3, &tmp2) != MP_OKAY) { goto error; }
   if (mp_sub(&tmp, &tmp2, &tmp) != MP_OKAY) { goto error; }
   if (mp_add(&tmp, &p, &tmp) != MP_OKAY) { goto error; }
   if (mp_read_radix(&p, sets[idx].prime, 10) != MP_OKAY) { goto error; }
   if (mp_mod(&tmp, &p, &tmp) != MP_OKAY) { goto error; }

   /* now find square root */
   if (mp_add_d(&p, 1, &tmp2) != MP_OKAY) { goto error; }
   if (mp_div_2(&tmp2, &tmp2) != MP_OKAY) { goto error; }
   if (mp_div_2(&tmp2, &tmp2) != MP_OKAY) { goto error; }              /* tmp2 = (p+1)/4 */
   if (mp_exptmod(&tmp, &tmp2, &p, &tmp) != MP_OKAY) { goto error; }   /* tmp  = (x^3 - 3x + b)^((p+1)/4) mod p */

   /* if result==0, then y==tmp, otherwise y==p-tmp */
   if (result == 0) {
      if (mp_copy(&tmp, &pt->y) != MP_OKAY) { goto error; }
   } else {
      if (mp_sub(&p, &tmp, &pt->y) != MP_OKAY) { goto error; }
   }
   
   res = CRYPT_OK;
   goto done;
error:
   res = CRYPT_ERROR;
   //crypt_error = "Out of memory in expand_y_point().";
done:
   mp_clear_multi(&p, &tmp, &tmp2, NULL);
   return res;
}
#endif

#define OUTPUT_BIGNUM(num, buf2, y, z)         \
{                                              \
      z = mp_raw_size(num);                    \
      STORE32L(z, buf2+y);                     \
      y += 4;                                  \
      mp_toraw(num, buf2+y);                   \
      y += z;                                  \
}


#define INPUT_BIGNUM(num, in, x, y)                              \
{                                                                \
     /* load value */                                            \
     LOAD32L(x, in+y);                                           \
     y += 4;                                                     \
                                                                 \
     /* sanity check... */                                       \
     if (x > 1024) {                                             \
        /*crypt_error = "Invalid size of data in ecc_import().";*/   \
        goto error;                                              \
     }                                                           \
                                                                 \
     /* load it */                                               \
     if (mp_read_raw(num, (unsigned char *)in+y, x) != MP_OKAY) {\
        /*crypt_error = "Out of memory in ecc_import().";*/          \
        goto error;                                              \
     }                                                           \
     y += x;                                                     \
}

#if 0
int my_ecc_export(unsigned char *out, unsigned char *outlen, int type, ecc_key *key)
{
   unsigned char y, z;
   int res;

   /* output type and magic byte */
   y = PACKET_SIZE;
   out[y++] = type;
   out[y++] = 0; //key->idx;
//   dbsp("EXP_1 ");dbch(y);dbc('\n');

   /* output x coordinate */
   OUTPUT_BIGNUM(&key->pubkey.x, out, y, z);
//   dbsp("EXP_2 ");dbch(y);dbc('\n');

   /* compress y and output it  */
   if (compress_y_point(&key->pubkey, /*key->idx, */ &res) == CRYPT_ERROR) {
//   dbsp("EXP_3 ");dbch(y);dbc('\n');
      return CRYPT_ERROR;
   }
   out[y++] = res;
//   dbsp("EXP_4 ");dbch(y);dbc('\n');

   if (type == PK_PRIVATE) {
      OUTPUT_BIGNUM(&key->k, out, y, z);
//   dbsp("EXP_5 ");dbch(y);dbc('\n');
   }

   *outlen = y;
//   dbsp("EXP_6 ");dbch(y);dbc('\n');
   return CRYPT_OK;
}

#if 0
int ecc_import(const unsigned char *in, ecc_key *key)
{
   unsigned long x, y;
   int res;

   /* check type */
   /*
   if (packet_valid_header((unsigned char *)in, PACKET_SECT_ECC, PACKET_SUB_KEY) == CRYPT_ERROR) { 
      return CRYPT_ERROR;
   }
   */

   /* init key */
   if (mp_init_multi(&key->pubkey.x, &key->pubkey.y, &key->k, NULL) != MP_OKAY) {
      //crypt_error = "Out of memory in ecc_import().";
      return CRYPT_ERROR;
   }

   y = PACKET_SIZE;
   key->type = in[y++];
   key->idx  = in[y++];

   /* type check both values */
   /*
   if ((key->type != PK_PUBLIC) && (key->type != PK_PRIVATE))  {
      //crypt_error = "Invalid type of key as input for ecc_import().";
      goto error;
   }
   */
   
   /* is the key idx valid? */
   /*
   if (!is_valid_idx(key->idx)) {
      //crypt_error = "Invalid key idx found in input for ecc_import().";
      goto error;
   }
   */
   /* load x coordinate */
   INPUT_BIGNUM(&key->pubkey.x, in, x, y);
  
   /* load y */
   x = in[y++];
   if (expand_y_point(&key->pubkey, key->idx, x) == CRYPT_ERROR) { goto error; }

   if (key->type == PK_PRIVATE) {
      /* load private key */
      INPUT_BIGNUM(&key->k, in, x, y);
   }
   res = CRYPT_OK;
   goto done;
error:
   res = CRYPT_ERROR;
   mp_clear_multi(&key->pubkey.x, &key->pubkey.y, &key->k, NULL);
done:
   return res;
}
#endif

/* Signatures 
 *
 * Signatures are performed using a slightly modified ElGamal protocol.  
 * In these notes uppercase letters are points and lowercase letters are 
 * scalars.  The users private key is 'x' and public key is Y = xG.  
 * The order of the curve is 'r'.
 *
 *
 * To sign a message 'm' the user does this

1.  Makes up a random 'k' and finds kG [basically makes up a ecc_key], we will let A = kG
2.  Finds b such that b = (m - x)/k mod r
3.  Outputs (A, b) as the signature

To verify a user computes mG and compares that to (bA + Y).  Note that (bA + Y) is equal to

= ((m - x)/k)(kG) + xG
= (m - x)G + xG
= mG

In theory, assuming the ECC Discrete Log is a hard problem an attacker 
cannot find 'x' from (A, b).  'b' is perfectly decorrelated and reveals no 
information.  A reveals what kG is but not 'k' directly.  Therefore, 
assuming finding 'k' given kG is hard, finding 'x' from b is hard too.

 * Message format for a message signature

offset    | size  |  description
---------------------------------
   0      |   1   | hash ID
   1      |   4   | size of public key (A)
   5      |   b   | public key (A)
   5+b    |   4   | size of mp_int (b)
   9+b    |   c   | mp_int (b)
---------------------------------
*/

#define mp_out_8 out_mp_8

int my_ecc_sign(ecc_key *key, ecc_key * pubkey, unsigned char *md, unsigned char * out, unsigned char *outlen)
{
   mp_int b, p;
   unsigned char epubkey[100];
   unsigned char x, y, z, pubkeysize, rsize;
   int res;

   /*
   dbsp("SIGN\n");
   dbsp("K_K\n");out_mp_8(&key->k);
   dbsp("K_X\n");out_mp_8(&key->pubkey.x);
   dbsp("K_Y\n");out_mp_8(&key->pubkey.y);

   dbsp("PK_K\n");out_mp_8(&pubkey->k);
   dbsp("PK_X\n");out_mp_8(&pubkey->pubkey.x);
   dbsp("PK_Y\n");out_mp_8(&pubkey->pubkey.y);
   */

   pubkeysize = sizeof(epubkey);
   if (my_ecc_export(epubkey, &pubkeysize, PK_PUBLIC, pubkey) == CRYPT_ERROR) {
      return CRYPT_ERROR;
   }

   y = PACKET_SIZE;
   out[y++] = 0xAA;
   out[y++] = pubkeysize;
//   STORE32L(pubkeysize, out+y);
   y += 3;
   for (x = 0; x < pubkeysize; x++, y++) {
       out[y] = epubkey[x];
   }

   if (mp_init_multi(&b, &p, NULL) != MP_OKAY) { 
      return CRYPT_ERROR;
   }

   if (mp_read_radix(&p, ecc192_order, 10) != MP_OKAY)            { goto error; }
   if (mp_read_raw(&b, md, 1+20) != MP_OKAY)  { goto error; }
//   mp_out_8(&b);

   /* find b = (m - x)/k */
   if (mp_invmod(&pubkey->k, &p, &pubkey->k) != MP_OKAY) { goto error; } /* k = 1/k */
   if (mp_submod(&b, &key->k, &p, &b) != MP_OKAY) { goto error; }      /* b = m - x */
   if (mp_mulmod(&b, &pubkey->k, &p, &b) != MP_OKAY) { goto error; }    /* b = (m - x)/k */

   rsize = mp_raw_size(&b);
   if (rsize > sizeof(epubkey)) { 
      goto error; 
   }
   mp_toraw(&b, epubkey);
   
//   STORE32L(rsize, out+y);
   out[y++]=rsize;
   y += 3;
   for (x = 0; x < rsize; x++, y++) {
       out[y] = epubkey[x];
   }

   *outlen = y;

   res = CRYPT_OK;
   goto done1;
error:
   res = CRYPT_ERROR;
done1:
   mp_clear_multi(&b, &p, NULL);
   return res;   
}
#endif
